
/*
	================================================================================
	Library for composing and rendering 3D scenes (move/rotate objects/camera)
	Copyright 1994-2002 (c) Ano-Tech Computers -- All Rights Reserved
	Written by Andy Floyd with lots of help from Allman

    Modify and distribute freely as long as this header is kept intact and
	the final program credits Ano-Tech Computers

    IMPORTANT: Call init_math3d before attempting to use any of the functions
    IMPORTANT: Call shutdown_math3d before exiting or your program will leak memory
	================================================================================
*/



#include "raster.h"

// Compile-time limits - these affect the possible scene complexity,
// higher values consume more memory
#define MAX_OBJECTS 32		// In memory
#define MAX_MODELS 16		// In memory
#define MAX_VERTICES 102400	// Total for all _objects_ with flags & O3D_RENDER
#define MAX_SURFACES 10240	// Total for all _objects_ with flags & O3D_RENDER

#define VERTICES_PER_SURFACE 16	// Used when rendering only, adjust this value
								// if you have more complex surfaces.
								// Must be <= MAX_VERTICES 1024

#define MODEL_VERSION_REQUIRED 1	// All .m3d files must match this version number!

// object3d flags - these flags are used when calculating object3d[].flags
#define O3D_EMPTY 1			// Object slot is empty
#define O3D_RENDER 2		// Object is visible and should be rendered

#define CIRCLE 7200		// Circles are divided into this many units
						// A higher number gives greater precision in rotation but
						// also higher memory consumption for precalc tables.
						// Lower than 360 will produce very ugly results...

#define LENS 250   		// This is the DEFAULT perspective setting, camera.lens
						// can be adjusted runtime between each frame.
						// A shorter lens gives a larger Field of View
						// with more distorted geometry ("fisheye")
						// A longer lens means a smaller Field of View with
						// less distorted geometry ("zoom")
						// Recommended setting is appx. HEIGHT/2

#define AMBIENT 32		// Used for shading, this is the minimum amount of light present
						// (It will be added to the shading light source)
						// Range is 0 - 255, recommended value is < 64

// ------------------------- no configuration below this line -------------------------

// sin/cos precalc tables populated by precalc_trigonometry() and referenced 
// by the 3D code.
static int Sine[CIRCLE];
static int Cosine[CIRCLE];


typedef struct {
	int x;
	int y;
	int z;
	unsigned int color;
} vertex3d;

typedef struct {
    // Each surface points to a list of index numbers referring to model3d[].vertexlist
	unsigned int *vertexlist;
	unsigned int vertices;
	unsigned int color;
	int texture;
	unsigned u;
	unsigned v;
	unsigned int first_vertex;	// Set dynamically by compose_scene()
} surface3d;

typedef struct {
	char filename[260];
	unsigned int vertices;
	vertex3d *vertexlist;
	unsigned int surfaces;
	surface3d *surfacelist;
} model3d;

typedef struct {
	int x;		// Keep in mind that coordinates below < 10 
	int y;		//		may "jitter" pretty bad when rotated close up
	int z;		//		because rotation is done with integer math
				//		Max. value is really determined by many factors but
				//		we have no problem with values +/- 100000
	int pitch;	// Rotation arounx X axis
	int	yaw;	// Rotation around Y axis
	int roll;	// Rotation around Z axis
	short int model;	// math3d_model[] index number
	unsigned int flags;
} object3d;

typedef struct {
	int x;
	int y;
	int z;
	int pitch;	// Rotation arounx X axis
	int	yaw;	// Rotation around Y axis
	int roll;	// Rotation around Z axis
	int lens;
	int depth;
} camera3d;

// The camera may be adjusted freely, changes affect the next render_* call
// One thing worth mentioning: camera is rotated around its own axis, not
// the scene or the focal point. (i.e. rotating the camera is like rotating your head)
// This seems like the most intuitive way.
camera3d	camera;

// This version uses short integers to reference objects
// THIS MAY CHANGE IN THE FUTURE -- TREAT THEM AS HANDLES!!
typedef short int math3d_object; 
typedef short int math3d_model; 


// This version uses static arrays internally
// THIS MAY CHANGE IN THE FUTURE -- DO NOT ACCESS DIRECTLY!!
object3d	math3d_objects	[MAX_OBJECTS ];
model3d		math3d_models	[MAX_MODELS  ];	
vertex3d	math3d_vertices	[MAX_VERTICES];
surface3d	math3d_surfaces	[MAX_SURFACES];
unsigned int vertices_in_scene = 0;
unsigned int surfaces_in_scene = 0;


int hextoint(char *string)
{
	// Convert a string in the format '0xFFFFFFFF' to its integer equivalent
	// Supports variable length strings
	// The first two chars are assumed to be '0x' and silently SKIPPED
	char digits[16] = "0123456789ABCDEF";
	unsigned int pos;
	unsigned int index;
	unsigned int scan;
	int result = 0;
	for (pos=2; pos<strlen(string); pos++)
	{
		result = result * 16;
		index = 0;
		for (scan=0; scan<16; scan++)
		{
			if (digits[scan] == string[pos]) { index = scan; }
		}
		result = result + index;
	}
	return (result);
}

void precalc_trigonometry()
{
	// Populates the global tables Sine and Cosine with precalced values for
	// a fill circle divided into 8192. The trigonometrical value is multiplied
	// by 8192 and stored as an integer to achieve very fast 3D calculation
	// Example lookup where degrees is 0-359: Sine[(degrees * CIRCLE) / 360]
	double pi = 3.1415926535;
	double radians;
	int index;

	for (index=0; index<CIRCLE; index++)
	{
		radians = (pi*2 * index) / CIRCLE;
		Sine[index] = (int) ((sin(radians) * CIRCLE) + 0.5);
		Cosine[index] = (int) ((cos(radians) * CIRCLE) + 0.5);
	}
}


int fast_sin(int units)
{
	// Look up the precalculated trigonometrical sin() value for an angle
	// described by 'units', where a full circle is 'CIRCLE' units.
	// The value can bee looked up directly from Sine[] but this function
	// handles units outside the 0 to CIRCLE-1 range correctly.
	// Note: The returned value must be divided by CIRCLE to make sense
	units = units % CIRCLE;
	if (units < 0) { units = CIRCLE + units; }	// negative -> positive
	return (Sine[units]);
}


int fast_cos(int units)
{
	// Look up the precalculated trigonometrical cos() value for an angle
	// described by 'units', where a full circle is 'CIRCLE' units.
	// The value can bee looked up directly from Cosine[] but this function
	// handles units outside the 0 to CIRCLE-1 range correctly.
	// Note: The returned value must be divided by CIRCLE to make sense
	units = units % CIRCLE;
	if (units < 0) { units = CIRCLE + units; }	// negative -> positive
	return (Cosine[units]);
}



math3d_model load_model3d(char *modelname)
{
	// Read and parse the file, and prepare a model3d structure
	// Returns a pointer to the model3d structure if successful
	// A model can NOT be plotted, it is only a "blueprint"
	// used to create objects.
	FILE *filehandle;
	char filename[260];
//	char *path = "release\\resource\\models\\";
	char *path = "";
	int index;
	unsigned int filesize;
	unsigned int bytesread;
	char *filebuffer;
	unsigned char comment = 0;	// 1 when between # and \n
	math3d_model model = -1;
	unsigned char loaded = 0;
	unsigned int pos = 0;
	int state = 0;			// 0 = looking for type, version, vertices, surfaces
	char word[256];
	int abort = 0;			// Set to 1 if a fatal parse error occurs
	int wordlen;
	int file_type = 0;
	int file_version = 0;
	int file_vertices = 0;
	int file_surfaces = 0;
	vertex3d *vertexlist;
	surface3d *surfacelist;
	int handled = 0;
	unsigned int subpos;
	char subword[16];	// No vertex number is this long
	unsigned int * vertices;
	unsigned int vertexcount;

	char debug[256];

	// Find an unused slot in the math3d_models[] table
	for (index=0; index<MAX_MODELS; index++)
	{
		if (strcmp(math3d_models[index].filename, "") == 0)
		{
			model = index;
			break;
		}
	}

	// "model" now contains a slot number OR -1 if none are available
	if (model > -1) 
	{
		// Open the model file for read
		filename[0] = 0;			// filename = ""
		strcat(filename, path);		// filename .= path
		strcat(filename, modelname);		// filename .= fontname
		filehandle = fopen(filename, "rb");
		if (filehandle != NULL)
		{
			// Determine the file size
			fseek(filehandle, 0, SEEK_END);		// Seek to the end...
			filesize = ftell(filehandle);		// ...to find out how big the file is
			rewind(filehandle);					// Rewind so we can read the file
			if (filesize > 0) 
			{
				// Allocate a buffer
				filebuffer = (unsigned char *) malloc(filesize);
				if (filebuffer != NULL) 
				{
					// Read file into buffer
					bytesread = fread(filebuffer, 1, filesize, filehandle);
					if (bytesread == filesize)
					{
						// Parse the file contents
						loaded = 1;
						strcpy(math3d_models[model].filename, filename);
						math3d_models[model].vertices = 0;
						math3d_models[model].surfaces = 0;
						strcpy(word, "");
						
						while (pos < filesize && abort == 0) 
						{
							handled = 0;
							if (filebuffer[pos] == 35) // # starts a comment
							{ 
								comment = 1; 
							}	

							if (filebuffer[pos] == 13) // \r = \s
							{ 
								filebuffer[pos] = 32; 
							}	

							if (filebuffer[pos] == 9) // \t = \s
							{ 
								filebuffer[pos] = 32; 
							}	

							if (filebuffer[pos] == 10) // \n ends a comment
							{ 
								comment = 0;
								filebuffer[pos] = 32;	// Treat as a space
							}

							// Skip other tests if we are inside a comment
							if (comment == 0) 
							{
								if (filebuffer[pos] == 32)
								{
									/*strcpy(debug, "|");
									strcat(debug, word);
									strcat(debug, "|\n");
									plot_string(debug, -1,-1);*/

									if (state == 9)	// looking for a vertex x value
									{
										if (strlen(word) > 0)
										{
											vertexlist = math3d_models[model].vertexlist;
											vertexlist[index].x = atoi(word);
											state = 8;
											handled = 1;
										}
									}

									if (state == 10)	// looking for a vertex y value
									{
										if (strlen(word) > 0)
										{
											vertexlist = math3d_models[model].vertexlist;
											vertexlist[index].y = atoi(word);
											state = 8;
											handled = 1;
										}
									}

									if (state == 11)	// looking for a vertex z value
									{
										if (strlen(word) > 0)
										{
											vertexlist = math3d_models[model].vertexlist;
											vertexlist[index].z = atoi(word);
											state = 8;
											handled = 1;
										}
									}

									if (state == 12)	// looking for a vertex color
									{
										if (strlen(word) > 0)
										{
											vertexlist = math3d_models[model].vertexlist;
											vertexlist[index].color = hextoint(word);
											state = 8;
											handled = 1;
										}
									}

									if (state == 8 && handled == 0)	// looking for a vertex detail OR a closing bracket
									{
										if (strcmp(word, "x") == 0) { state = 9; }
										if (strcmp(word, "y") == 0) { state = 10; }
										if (strcmp(word, "z") == 0) { state = 11; }
										if (strcmp(word, "color") == 0) { state = 12; }
										if (strcmp(word, "colour") == 0) { state = 12; }
										// state numbers 13-19 reserved
										if (strcmp(word, "}") == 0) { state = 6; }
										handled = 1;
									}

									if (state == 7)	// looking for an opening vertex bracket
									{
										if (strcmp(word, "{") == 0) 
										{ 
											state = 8; 
											vertexlist = math3d_models[model].vertexlist;
											vertexlist[index].color = WHITE;

										}
									}

									if (state == 6 && handled == 0)	// looking for a vertex no OR a closing bracket
									{
										if (strcmp(word, "}") == 0) 
										{ 
											state = 0; 
										}	
										else
										{
											if (strlen(word) > 0) 
											{
												index = atoi(word);
												state = 7;
											}
										}
									}

									if (state == 5)	// looking for an opening vertices bracket
									{
										if (strcmp(word, "{") == 0) { state = 6; }
									}

									if (state == 1)	// looking for type string
									{
										state = 0;
										file_type = 1;
										if (strcmp(word, "model3d") != 0) 
										{ 
											plot_string("ERROR: Incorrect file type\n", -1,-1);
											abort = 1; 
										}	
									}

									if (state == 2)	// looking for version number
									{
										state = 0;
										file_version = atoi(word);

										if (file_version != MODEL_VERSION_REQUIRED) 
										{
											plot_string("ERROR: Incorrect file version\n", -1,-1);
											abort = 1; 
										}	
									}

									if (state == 3)	// looking for a vertex count
									{
										state = 5;
										file_vertices = atoi(word);

										if (file_vertices > 0) 
										{
											math3d_models[model].vertices = file_vertices;
											math3d_models[model].vertexlist = (vertex3d *) malloc(sizeof(vertex3d) * file_vertices);
										}
										else
										{
											plot_string("ERROR: Minimum 1 vertex required\n", -1,-1);
											abort = 1; 
										}
									}

									if (state == 25)	// looking for a surface vertex list
									{
										if (strlen(word) > 0)
										{
											surfacelist = math3d_models[model].surfacelist;
											surfacelist[index].vertices = 0;
											strcat(word, ",");	// Add a trailing comma for easier parsing
											vertexcount = 0;
											// Count the number of vertices for this surface
											for (subpos=0; subpos<strlen(word); subpos++)
											{
												if (word[subpos] == 44) { vertexcount++; }
											}

											// Get each vertex and put the number into our vertex index
											vertices = (unsigned int *) malloc(sizeof(unsigned int) * vertexcount);
											surfacelist[index].vertexlist = vertices;
											surfacelist[index].vertices = vertexcount;
											vertexcount = 0;
											strcpy(subword, "");
											for (subpos=0; subpos<strlen(word); subpos++)
											{
												if (word[subpos] == 44)
												{
													//sprintf(debug, "Surface %i vertex %i is %i\n", index, vertexcount, atoi(subword));
													//plot_string(debug, -1,-1);
													vertices[vertexcount] = atoi(subword);
													vertexcount++;
													strcpy(subword, "");											
												}
												else
												{
													wordlen = strlen(subword);
													subword[wordlen] = word[subpos];
													subword[wordlen+1] = '\0';	// Or just plain 0
												}
											}
											
											state = 24;
											handled = 1;
										}
									}

									if (state == 26)	// looking for a surface color
									{
										if (strlen(word) > 0)
										{
											surfacelist = math3d_models[model].surfacelist;

											surfacelist[index].color = hextoint(word);
											state = 24;
											handled = 1;
										}
									}

									if (state == 27)	// looking for a texture filename
									{
										if (strlen(word) > 0)
										{
											surfacelist[index].texture = load_bitmap(word);
											state = 24;
											handled = 1;
										}
									}
									if (state == 24 && handled == 0)	// looking for a surface detail OR a closing bracket
									{
										if (strcmp(word, "vertices") == 0) { state = 25; }
										if (strcmp(word, "color") == 0) { state = 26; }
										if (strcmp(word, "colour") == 0) { state = 26; }
										if (strcmp(word, "texture") == 0) { state = 27; }
										// state numbers 28-99 reserved
										if (strcmp(word, "}") == 0) 
										{ 
											index++; 
											state = 22; 
										}
										handled = 1;
									}

									if (state == 23)	// looking for an opening surface bracket
									{
										if (strcmp(word, "{") == 0) { 
											state = 24; 
											// Assign default parameters for this surface
											surfacelist = math3d_models[model].surfacelist;
											surfacelist[index].texture = -1;
											surfacelist[index].color = WHITE;
											surfacelist[index].vertices = 0;
										}
									}

									if (state == 22 && handled == 0)	// looking for a surface name/no OR a closing bracket
									{
										if (strcmp(word, "}") == 0) 
										{ 
											state = 0; 
										}	
										else
										{
											if (strlen(word) > 0) 
											{
												state = 23;
											}
										}
									}


									if (state == 21)	// looking for an opening surfaces bracket
									{
										if (strcmp(word, "{") == 0) 
										{ 
											state = 22; 
											index = 0;
										}
									}

									if (state == 20)	// looking for a surface count
									{
										state = 21;
										file_surfaces = atoi(word);

										if (file_surfaces > 0) 
										{
											math3d_models[model].surfaces = file_surfaces;
											math3d_models[model].surfacelist = (surface3d *) malloc(sizeof(surface3d) * file_surfaces);
										}
									}

									if (state == 0 && handled == 0)	// looking for type, version, vertices, surfaces
									{
										if (strcmp(word, "type") == 0) { state = 1; }
										if (strcmp(word, "version") == 0) { state = 2; }
										if (strcmp(word, "vertices") == 0) { state = 3; }
										if (strcmp(word, "surfaces") == 0) { state = 20; }
									}

									strcpy(word, "");
								}
								else
								{
									// Add this character to the word buffer
									wordlen = strlen(word);
									word[wordlen] = filebuffer[pos];
									word[wordlen+1] = '\0';	// Or just plain 0
								}
							}

							pos++;
						}

						// De-allocate the buffer
						free(filebuffer);
					}
				}
			}
			
			fclose(filehandle);
		}
	}
	
	return (model);
}



math3d_object create_object3d(math3d_model model)
{
	// Takes a math3d_model handle as parameter..
	// Creates an object based on that model
	// Returns the math3d_object handle if successful
	// Returns -1 if an error occurred
	math3d_object index;
	math3d_object object = -1;

	// Verify that model is valid
	if (strcmp(math3d_models[model].filename, "") == 0) { return NULL; }

	// Find an unused slot in the math3d_objects[] table
	for (index=0; index<MAX_OBJECTS; index++)
	{
		if (math3d_objects[index].flags & O3D_EMPTY)
		{
			object = index;
			break;
		}
	}


	// "object" now contains a slot number OR -1 if none are available
	if (object > -1) 
	{
		math3d_objects[object].model = model;
		math3d_objects[object].x = 0;
		math3d_objects[object].y = 0;
		math3d_objects[object].z = 0;
		math3d_objects[object].pitch = 0;
		math3d_objects[object].yaw = 0;
		math3d_objects[object].roll = 0;
		math3d_objects[object].flags = O3D_RENDER;
	}

	return (object);
}




void init_math3d()
{
	// Do initialization stuff
	math3d_object object;
	math3d_model model;
	precalc_trigonometry();
	init_raster();

	// Init object table
	for (object=0; object<MAX_OBJECTS; object++)
	{
		math3d_objects[object].flags = O3D_EMPTY;
	}

	// Init model table
	for (model=0; model<MAX_MODELS; model++)
	{
		strcpy(math3d_models[model].filename, "");	// = unused
		math3d_models[model].vertices = 0;
		math3d_models[model].surfaces = 0;
	}

	// Place camera
	camera.x = 0;
	camera.y = 0;
	camera.z = 0;
	camera.pitch = 0;
	camera.yaw   = 0;
	camera.roll  = 0;
	camera.lens  = LENS;
	camera.depth = 100;
}



void destroy_model3d(math3d_model model)
{
	// Destroy the specified 3D model and all its sub components (if any)
	unsigned int index;
	surface3d * surfacelist;
	surface3d surface;
	unsigned int * vertexlist;

	// Destroy the vertex list for each surface
	for (index=0; index<math3d_models[model].surfaces; index++)
	{
		surfacelist = math3d_models[model].surfacelist;

		if (surfacelist != NULL) 
		{
			surface = *(surfacelist + index);
			vertexlist = surface.vertexlist;
			if (vertexlist != NULL)
			{
				free(vertexlist);
			}
		}
	}

	// Destroy the surface list itself
	if (math3d_models[model].surfacelist != NULL) 
	{ 
		free(math3d_models[model].surfacelist); 
	}

	// Destroy the vertex list
	if (math3d_models[model].vertexlist != NULL) 
	{ 
		free(math3d_models[model].vertexlist); 
	}
	
	strcpy(math3d_models[model].filename, "");	// = unused
	math3d_models[model].vertices = 0;
	math3d_models[model].surfaces = 0;
}



void destroy_object3d(math3d_object object)
{
	// Destroy the specified 3D object and all its sub components (if any)
	math3d_objects[object].flags = O3D_EMPTY;
}


void shutdown_math3d()
{
	// Deallocate all data structures used by math3d
	math3d_object object;
	math3d_model model;

	// Destroy 3D objects
	for (object=0; object<MAX_OBJECTS; object++)
	{
		destroy_object3d(object);
	}

	// Destroy 3D models
	for (model=0; model<MAX_MODELS; model++)
	{
		destroy_model3d(model);
	}

}



math3d_model create_testmodel3d(int size)
{
	// Returns a handle to a test model as if it was loaded from disk
	int index;
	int count;
	math3d_model model = -1;
	vertex3d * vertexlist;
	surface3d * surfacelist;
	unsigned int * vertices;

	// Find an unused slot in the math3d_models[] table
	vertexlist = NULL;
	for (index=0; index<MAX_MODELS; index++)
	{
		if (! strcmp(math3d_models[index].filename, ""))
		{
			model = index;
			break;
		}
	}



	// "model" now contains a slot number OR -1 if none are available
	if (model > -1) 
	{
		strcpy(math3d_models[model].filename, "(test)");

		// Insert vertices
		count = 8;
 		vertexlist = (vertex3d *) malloc(sizeof(vertex3d) * count);
		if (vertexlist != NULL)
		{
			// Define the vertices
			math3d_models[model].vertices = count;
			math3d_models[model].vertexlist = vertexlist;

			vertexlist[0].x = -size;
			vertexlist[0].y = -size;
			vertexlist[0].z = -size;

			vertexlist[1].x =  size;
			vertexlist[1].y = -size;
			vertexlist[1].z = -size;

			vertexlist[2].x = -size;
			vertexlist[2].y =  size;
			vertexlist[2].z = -size;

			vertexlist[3].x =  size;
			vertexlist[3].y =  size;
			vertexlist[3].z = -size;

			vertexlist[4].x = -size;
			vertexlist[4].y = -size;
			vertexlist[4].z =  size;

			vertexlist[5].x =  size;
			vertexlist[5].y = -size;
			vertexlist[5].z =  size;

			vertexlist[6].x = -size;
			vertexlist[6].y =  size;
			vertexlist[6].z =  size;

			vertexlist[7].x =  size;
			vertexlist[7].y =  size;
			vertexlist[7].z =  size;

			// Insert surfaces
			count = 6;
 			surfacelist = (surface3d *) malloc(sizeof(surface3d) * count);
			if (surfacelist != NULL)
			{
				// Define the vertices and surfaces
				math3d_models[model].surfaces = count;
				math3d_models[model].surfacelist = surfacelist;

				// Front
				vertices = (unsigned int *) malloc(sizeof(unsigned int) * 4);
				vertices[0] = 0;
				vertices[1] = 2;
				vertices[2] = 3;
				vertices[3] = 1;
				surfacelist[0].vertexlist = vertices;
				surfacelist[0].vertices = 4;
				surfacelist[0].texture = -1;
				surfacelist[0].color = 0xFF0000;

				// Back
				vertices = (unsigned int *) malloc(sizeof(unsigned int) * 4);
				vertices[0] = 4;
				vertices[1] = 5;
				vertices[2] = 7;
				vertices[3] = 6;
				surfacelist[1].vertexlist = vertices;
				surfacelist[1].vertices = 4;
				surfacelist[1].texture = -1;
				surfacelist[1].color = 0x7F0000;

				// Top
				vertices = (unsigned int *) malloc(sizeof(unsigned int) * 4);
				vertices[0] = 0;
				vertices[1] = 4;
				vertices[2] = 6;
				vertices[3] = 2;
				surfacelist[2].vertexlist = vertices;
				surfacelist[2].vertices = 4;
				surfacelist[2].texture = -1;
				surfacelist[2].color = 0x00FF00;

				// Bottom
				vertices = (unsigned int *) malloc(sizeof(unsigned int) * 4);
				vertices[0] = 5;
				vertices[1] = 1;
				vertices[2] = 3;
				vertices[3] = 7;
				surfacelist[3].vertexlist = vertices;
				surfacelist[3].vertices = 4;
				surfacelist[3].texture = -1;
				surfacelist[3].color = 0x007F00;

				// Left
				vertices = (unsigned int *) malloc(sizeof(unsigned int) * 4);
				vertices[0] = 4;
				vertices[1] = 0;
				vertices[2] = 1;
				vertices[3] = 5;
				surfacelist[4].vertexlist = vertices;
				surfacelist[4].vertices = 4;
				surfacelist[4].texture = -1;
				surfacelist[4].color = 0x0000FF;

				// Right
				vertices = (unsigned int *) malloc(sizeof(unsigned int) * 4);
				vertices[0] = 2;
				vertices[1] = 6;
				vertices[2] = 7;
				vertices[3] = 3;
				surfacelist[5].vertexlist = vertices;
				surfacelist[5].vertices = 4;
				surfacelist[5].texture = -1;
				surfacelist[5].color = 0x00007F;
			}
		}

    }
	
	return (model);
}


void compose_scene()
{
	// Copy all vertices into the scene, moving and rotating them as needed
	// This is where we actually USE all the coordinates and angles
	// After this is done, we can render the scene in multiple ways
	// using the render_* functions
	math3d_object object;
	math3d_model model;
	vertex3d * vertexlist;
	unsigned int vertex;
	surface3d * surfacelist;
	unsigned int surface;
	int x;
	int y;
	int z;
	int xr;	// Temp variables for rotation
	int yr;
	int zr;

	// Reset scene counters
	vertices_in_scene = 0;
	surfaces_in_scene = 0;

	// Handle one object at a time
	for (object=0; object<MAX_OBJECTS; object++)
	{
		if (math3d_objects[object].flags & O3D_RENDER)
		{
			// Include this object in the scene -- first the vertices
			model = math3d_objects[object].model;
			if (math3d_models[model].vertices > 0)
			{
				// Handle one vertex at a time
				vertexlist = math3d_models[model].vertexlist;
				for (vertex=0; vertex<math3d_models[model].vertices; vertex++) 
				{
					// Get model vertex
					x = vertexlist[vertex].x;
					y = vertexlist[vertex].y;
					z = vertexlist[vertex].z;

					// Apply object roll
					xr =      (x * fast_cos(math3d_objects[object].roll)) / CIRCLE;
					xr = xr - (y * fast_sin(math3d_objects[object].roll)) / CIRCLE;
					yr =      (y * fast_cos(math3d_objects[object].roll)) / CIRCLE;
					yr = yr + (x * fast_sin(math3d_objects[object].roll)) / CIRCLE;
					x = xr;
					y = yr;

					// Apply object pitch
					yr =      (y * fast_cos(math3d_objects[object].pitch)) / CIRCLE;
					yr = yr + (z * fast_sin(math3d_objects[object].pitch)) / CIRCLE;
					zr =      (z * fast_cos(math3d_objects[object].pitch)) / CIRCLE;
					zr = zr - (y * fast_sin(math3d_objects[object].pitch)) / CIRCLE;
					y = yr;
					z = zr;

					// Apply object yaw
					xr =      (x * fast_cos(math3d_objects[object].yaw)) / CIRCLE;
					xr = xr - (z * fast_sin(math3d_objects[object].yaw)) / CIRCLE;
					zr =      (z * fast_cos(math3d_objects[object].yaw)) / CIRCLE;
					zr = zr + (x * fast_sin(math3d_objects[object].yaw)) / CIRCLE;
					x = xr;
					z = zr;					

					// Translate according to object position
					x = x + math3d_objects[object].x;
					y = y + math3d_objects[object].y;
					z = z + math3d_objects[object].z;

					// Store in scene table
					math3d_vertices[vertices_in_scene + vertex].x = x;
					math3d_vertices[vertices_in_scene + vertex].y = y;
					math3d_vertices[vertices_in_scene + vertex].z = z;
					math3d_vertices[vertices_in_scene + vertex].color = vertexlist[vertex].color;
				}
			}

			// Now the surfaces
			if (math3d_models[model].surfaces > 0)
			{
				// Handle one surface at a time
				surfacelist = math3d_models[model].surfacelist;
				for (surface=0; surface<math3d_models[model].surfaces; surface++) 
				{
					math3d_surfaces[surfaces_in_scene + surface].first_vertex = vertices_in_scene;
					math3d_surfaces[surfaces_in_scene + surface].vertices = surfacelist[surface].vertices;
					math3d_surfaces[surfaces_in_scene + surface].vertexlist = surfacelist[surface].vertexlist;
					math3d_surfaces[surfaces_in_scene + surface].color = surfacelist[surface].color;
					math3d_surfaces[surfaces_in_scene + surface].texture = surfacelist[surface].texture;
				}
			}

			// Update the scene counters
			vertices_in_scene = vertices_in_scene + math3d_models[model].vertices;
			surfaces_in_scene = surfaces_in_scene + math3d_models[model].surfaces;
		}
	}
}


void adjust_for_camera()
{
	// Rotate and move the entire scene according to the
	// position and angle of the camera
	unsigned int vertex;
	int x;
	int y;
	int z;
	int xr;
	int yr;
	int zr;
	for (vertex=0; vertex<vertices_in_scene; vertex++)
	{
		// Before we can rotate, move the pixel according to the camera position
		x = math3d_vertices[vertex].x;
		y = math3d_vertices[vertex].y;
		z = math3d_vertices[vertex].z - camera.z;

		// Apply camera roll
		xr =      (x * fast_cos(camera.roll)) / CIRCLE;
		xr = xr - (y * fast_sin(camera.roll)) / CIRCLE;
		yr =      (y * fast_cos(camera.roll)) / CIRCLE;
		yr = yr + (x * fast_sin(camera.roll)) / CIRCLE;
		x = xr;
		y = yr;

		// Apply camera pitch
		yr =      (y * fast_cos(camera.pitch)) / CIRCLE;
		yr = yr + (z * fast_sin(camera.pitch)) / CIRCLE;
		zr =      (z * fast_cos(camera.pitch)) / CIRCLE;
		zr = zr - (y * fast_sin(camera.pitch)) / CIRCLE;
		y = yr;
		z = zr;

		// Apply camera yaw
		xr =      (x * fast_cos(camera.yaw)) / CIRCLE;
		xr = xr - (z * fast_sin(camera.yaw)) / CIRCLE;
		zr =      (z * fast_cos(camera.yaw)) / CIRCLE;
		zr = zr + (x * fast_sin(camera.yaw)) / CIRCLE;
		x = xr;
		z = zr;
					
		// Save the new vertex
		math3d_vertices[vertex].x = xr;
		math3d_vertices[vertex].y = yr;
		math3d_vertices[vertex].z = zr;
		
	}
}


int perspective_corrected(int camera_xy, int xy, int z)	// Actually, it works equally well for y
{
	int pxy;

	int delta_xy;
	int delta_z;


	delta_xy = xy - camera_xy;
	delta_z = z - camera.z;


	// Prevent division by zero
	if (abs(delta_z) < 1) { delta_z = 1; }

	pxy = (camera.lens * delta_xy) / delta_z;
	return (pxy);
}



   


void render_pixels()
{
	// Show all objects as pixels with perspective and depth clipping
	unsigned int vertex;
	unsigned int surface;
	unsigned int index;
	unsigned int * vertexlist;
	int x;
	int y;
	int z;
/*	int delta_x;
	int delta_y;
	int delta_z;*/
	int px;
	int py;
/*	int angle_x;
	int angle_y;*/

	int normal;		// z value of a vector perpendicular to a surface


	unsigned int points;
	int points_x[VERTICES_PER_SURFACE];	
	int points_y[VERTICES_PER_SURFACE];
	int points_z[VERTICES_PER_SURFACE];

	compose_scene();
	adjust_for_camera();

	// Plot the points as circles
	for (vertex=0; vertex<vertices_in_scene; vertex++) 
	{
		x = math3d_vertices[vertex].x;
		y = math3d_vertices[vertex].y;
		z = math3d_vertices[vertex].z;
		color = math3d_vertices[vertex].color;
		if (color == BLACK) { color = WHITE; }
		if (z > 0) {

			// Adjust for perspective
			px = perspective_corrected(camera.x, x, z);
			py = perspective_corrected(camera.y, y, z);
			plot_pixel(px + width/2, py + height/2);
		}
	}



}	// render_pixels()


int normal_z(x1, y1, z1, x2, y2, z2)
{
	// Calculate the normalized z vector of the normal of the two vectors passed
	// Get the normal vector of the two vectors passed
	int length;
	int x;
	int y;
	int z;

	x = (y1 * z2) - (z1 * y2);
	y = (z1 * x2) - (x1 * z2);
	z = (x1 * y2) - (y1 * x2);

	// Normalize and return a z value between -255 and 255
	length = (int) sqrt(x*x + y*y + z*z);
	if (length == 0) { length = 1; }
	
	return ((z * 255) / length);
}



int shaded_rgb(int rgb, unsigned int shade)
{
	// Split RGB components, apply shading info, then re-combine them
	int r;
	int g;
	int b;
	shade = shade * -1;					// Lower z means brighter color
	shade = shade + AMBIENT;
	if (shade > 255) { shade = 255; }

	// Copy B shade to R and G
//	shade = shade + (shade * 256) + (shade * 65536);

	r = rgb & 0xFF0000;
	r = (r * shade) / 256;
	r = r & 0xFF0000;
//	r = r + (shade & 0xFF0000);
//	r = (r >> 1) & 0xFF0000;

	g = rgb & 0x00FF00;
	g = (g * shade) / 256;
	g = g & 0x00FF00;
//	g = g + (shade & 0x00FF00);
//	g = (g >> 1) & 0x00FF00;

	b = rgb & 0x0000FF;
	b = (b * shade) / 256;
	b = b & 0x0000FF;
//	b = b + (shade & 0x0000FF);
//	b = (b >> 1) & 0x0000FF;


	return (r|g|b);
}


void render_depth()
{
	// Show all objects as filled polygons with perspective and depth clipping
	unsigned int vertex;
	unsigned int surface;
	unsigned int index;
	unsigned int * vertexlist;
	int x;
	int y;
	int z;
/*	int delta_x;
	int delta_y;
	int delta_z;*/
	int px;
	int py;
/*	int angle_x;
	int angle_y;*/
	int texture;

	int normal;		// z value of a vector perpendicular to a surface


	unsigned int points;
	int points_x[VERTICES_PER_SURFACE];	
	int points_y[VERTICES_PER_SURFACE];
	int points_z[VERTICES_PER_SURFACE];

	compose_scene();
	adjust_for_camera();

	// Plot the surface edges as lines to produce a wireframe
	for (surface=0; surface<surfaces_in_scene; surface++) 
	{
		points = math3d_surfaces[surface].vertices;

		color = math3d_surfaces[surface].color;
		vertexlist = math3d_surfaces[surface].vertexlist;
		texture = math3d_surfaces[surface].texture;
		for (vertex=0; vertex<points; vertex++) 
		{
			index = math3d_surfaces[surface].first_vertex + vertexlist[vertex];

			// Get vertex position
			x = math3d_vertices[index].x;
			y = math3d_vertices[index].y;
			z = math3d_vertices[index].z;


			// Adjust for perspective
			px = perspective_corrected(camera.x, x, z);
			py = perspective_corrected(camera.y, y, z);


			// Place vertex into scene
			points_x[vertex] = (width/2) - px;
			points_y[vertex] = (height/2) + py;
			points_z[vertex] = z - camera.z - camera.lens;

		}


		// Calculate the surface normal z value
//		normal = ((points_x[1]-points_x[0]) * (points_y[1]-points_y[2])) - ((points_y[1]-points_y[0]) * (points_x[1]-points_x[2]));
		normal = normal_z(points_x[1]-points_x[0], points_y[1]-points_y[0], points_z[1]-points_z[0], points_x[1]-points_x[2], points_y[1]-points_y[2], points_z[1]-points_z[2]);
		if (normal <= 0)
		{
			color = shaded_rgb(color, normal);
			fill_polygon3d(points, points_x, points_y, points_z, texture);
		}


	}	
	color = WHITE;

}	// render_flat()



void render_flat()
{
	// Show all objects as flat polysing with perspective, but with no depth clipping
	unsigned int vertex;
	unsigned int surface;
	unsigned int index;
	unsigned int * vertexlist;
	int x;
	int y;
	int z;
/*	int delta_x;
	int delta_y;
	int delta_z;*/
	int px;
	int py;
/*	int angle_x;
	int angle_y;*/

	int normal;		// z value of a vector perpendicular to a surface


	unsigned int points;
	int points_x[VERTICES_PER_SURFACE];	
	int points_y[VERTICES_PER_SURFACE];

	compose_scene();
	adjust_for_camera();

	// Plot the surface edges as lines to produce a wireframe
	for (surface=0; surface<surfaces_in_scene; surface++) 
	{
		points = math3d_surfaces[surface].vertices;

		color = math3d_surfaces[surface].color;
		vertexlist = math3d_surfaces[surface].vertexlist;
		for (vertex=0; vertex<points; vertex++) 
		{
			index = math3d_surfaces[surface].first_vertex + vertexlist[vertex];

			// Get vertex position
			x = math3d_vertices[index].x;
			y = math3d_vertices[index].y;
			z = math3d_vertices[index].z;


			// Adjust for perspective
			px = perspective_corrected(camera.x, x, z);
			py = perspective_corrected(camera.y, y, z);


			// Place vertex into scene
			points_x[vertex] = (width/2) - px;
			points_y[vertex] = (height/2) + py;

		}


		// Calculate the surface normal z value
		normal = ((points_x[1]-points_x[0]) * (points_y[1]-points_y[2])) - ((points_y[1]-points_y[0]) * (points_x[1]-points_x[2]));
		if (normal <= 0)
		{
			fill_polygon(points, points_x, points_y);
		}


	}	
	color = WHITE;

}	// render_flat()




void render_wireframe()
{
	// Show all objects as wireframes with perspective
	unsigned int vertex;
	unsigned int surface;
	unsigned int index;
	unsigned int * vertexlist;
	int x;
	int y;
	int z;
	int px;
	int py;
//	int angle_y;

	unsigned int points;
	int points_x[VERTICES_PER_SURFACE];	
	int points_y[VERTICES_PER_SURFACE];

	compose_scene();

	// Plot the surface edges as lines to produce a wireframe
	for (surface=0; surface<surfaces_in_scene; surface++) 
	{
		points = math3d_surfaces[surface].vertices;
		vertexlist = math3d_surfaces[surface].vertexlist;
		for (vertex=0; vertex<points; vertex++) 
		{
			index = math3d_surfaces[surface].first_vertex + vertexlist[vertex];

			// Get vertex position
			x = math3d_vertices[index].x;
			y = math3d_vertices[index].y;
			z = math3d_vertices[index].z;

			// Adjust for perspective
			px = perspective_corrected(camera.x, x, z);
			py = perspective_corrected(camera.y, y, z);

			// Place vertex into scene
			points_x[vertex] = (width/2) + px;
			points_y[vertex] = (height/2) + py;
		}
		plot_polygon(points, points_x, points_y);

	}


}	// render_wireframe()


void render_scene()
{
	// Show all objects with all bells and whistles
	compose_scene();

}	// render_scene()


void render_debug()
{
	// Show all objects as points and wireframes, without perspective or clipping
	unsigned int vertex;
	unsigned int surface;
	unsigned int index;
	unsigned int * vertexlist;
	int x;
	int y;
	int z;
	char debug[256];

	unsigned int points;
	int points_x[VERTICES_PER_SURFACE];	
	int points_y[VERTICES_PER_SURFACE];

	compose_scene();

	// Plot the points as circles
	for (vertex=0; vertex<vertices_in_scene; vertex++) 
	{
		sprintf(debug, "%i", vertex);
		x = math3d_vertices[vertex].x;
		y = math3d_vertices[vertex].y;
		z = math3d_vertices[vertex].z;
		color = math3d_vertices[vertex].color;
		plot_circle((width/2) + x, (height/2) + y, (z/20) + 10);
		color = WHITE;
		plot_string(debug, (width/2) + x + 10, (height/2) + y - 10);
	}

	// Plot the surface edges as lines to produce a wireframe
	for (surface=0; surface<surfaces_in_scene; surface++) 
	{
		points = math3d_surfaces[surface].vertices;

		vertexlist = math3d_surfaces[surface].vertexlist;
		for (vertex=0; vertex<points; vertex++) 
		{
			index = math3d_surfaces[surface].first_vertex + vertexlist[vertex];
			points_x[vertex] = (width/2) + math3d_vertices[index].x;
			points_y[vertex] = (height/2) + math3d_vertices[index].y;
		}
		plot_polygon(points, points_x, points_y);

	}

}



void move_object3d(math3d_object object, int x, int y, int z)
{
	// Move object to the absolute position specified
	if (math3d_objects[object].flags & O3D_EMPTY) { return; }
	math3d_objects[object].x = x;
	math3d_objects[object].y = y;
	math3d_objects[object].z = z;
	
}


void rotate_object3d(math3d_object object, int pitch, int yaw, int roll)
{
	// Set object rotation angles to the absolute angles specified
	// Note: Angles are given in units defined in CIRCLE
	if (math3d_objects[object].flags & O3D_EMPTY) { return; }
	math3d_objects[object].pitch = pitch;
	math3d_objects[object].yaw = yaw;
	math3d_objects[object].roll = roll;
}


void dolly_camera(int distance)
{
	// Move camera (distance) units in the direction it's currently pointing
	// (Much like an aircraft would do)
	// A negative distance moves the camera backwards
	// Camera direction is described by camera.pitch and camera.yaw
	camera.x = camera.x + ((distance * fast_sin(camera.yaw)) / CIRCLE);
	camera.y = camera.y + ((distance * fast_sin(camera.pitch)) / CIRCLE);
	camera.z = camera.z + ((distance * fast_cos(camera.yaw)) / CIRCLE);
}
