
/*
	================================================================================
	Raster library for plotting 2D primitives (lines, circles, polygons etc.)
	Copyright 1994-2002 (c) Ano-Tech Computers -- All Rights Reserved
	ASM written by Andy Floyd, some functions based on code written by The Joker
	C wrappers written by Andy Floyd with lots of help from Allman

    Modify and distribute freely as long as this header is kept intact and
	the final program credits Ano-Tech Computers

	NOTE: Always remember to set the public vars and calling init_raster()
	before calling any of the functions
	================================================================================
*/

#include "string.h"				// For strcat() and other string functions
#include "stdio.h"	// Can't remember why I included this... may be redundant

#define MAX_SCANLINES 480		// Used to determine buffer size
// NOTE: Must be >= the highest polygon to be plotted, usually = HEIGHT

#define MAX_CHAR_WIDTH 32		// Used when plotting ASCII characters
								// to prevent overflows due to corrupted font files

#define MAX_STRING_LENGTH 256	// Used when plotting ASCII strings
								// Adjust if you really need to plot longer strings.

// Pre-defined colors you may find useful when writing readable code
#define BLACK  0x000000
#define WHITE  0xFFFFFF
#define RED    0xFF0000
#define GREEN  0x00FF00
#define BLUE   0x0000FF
#define CYAN   0x00FFFF
#define PURPLE 0xFF00FF
#define YELLOW 0xFFFF00




// ------------------------- no configuration below this line -------------------------


// Raster primitives - public vars
static unsigned int *buffer;
static unsigned int color;
static unsigned int width;   
static unsigned int height;

// Raster primitives - private vars
static int PolyBuffer[MAX_SCANLINES * 2];
static unsigned int PolyBufferHi = MAX_SCANLINES - 1;
static unsigned int PolyBufferLo = 0;

static unsigned char *font_bitmap = NULL;
int font_bitmap_length = 0;
static unsigned char *font_index = NULL;
int font_index_length = 0;
static unsigned char *font_size = NULL;
int font_size_length = 0;

static int cursor_x = 1;	// Left margin
static int cursor_y = 1;	// Top of screen

static errorcode = 0;


#define MAX_BITMAPS 16		// In memory
typedef struct {
	unsigned int height;
	unsigned int width;
	char filename[260];
	unsigned int *bitmap;
} bitmap;
bitmap	raster_bitmaps	[MAX_BITMAPS];

typedef struct {
	int	x;
	int y;
	unsigned int color;
} vertex2d;


void init_raster()
{
	int index;
	for (index=0; index<MAX_BITMAPS; index++)
	{
		strcpy(raster_bitmaps[index].filename, "");
	}
}

void plot_pixel(int p1x, int p1y)
{
	// Plot a pixel at coordinated specified by parameters
	// Note: Pixel is clipped against top, bottom, left and right edge.
	// Note: None of my other raster functions actually uses this, it's
	// included in case anyone needs it.
	__asm {
		// p1x, p1y, color, buffer, width
		mov ebx,p1x
		cmp ebx,0
		jl  Clipped
		cmp ebx,width
		jge Clipped

		mov ecx,p1y
		cmp ecx,0
		jl  Clipped
		cmp ecx,height
		jge Clipped

		// ebx=X coord, ecx=Y coord, color, buffer, width
		mov eax,width	// eax = width
		mul ecx			// edx:eax = width * ecx
		add eax,ebx		// eax = eax + ebx

		// eax=pixel, color, buffer
		shl eax,2		// Each pixel is 4 bytes
		mov edi,buffer
		add edi,eax
		mov eax,color
		mov [edi],eax

			Clipped:
	}
}


void fill_buffer(unsigned int color)
{
	// Fill an entire buffer using the specified color
	// Use it to clear the screen or whatever
	// Input: color, buffer, height, width
	__asm {
		mov eax,height
		mul width	// edx:eax = eax * width
		mov ecx,eax
		mov eax,color
		mov edi,buffer
		rep stosd
	}
}


void plot_circle(int p1x, int p1y, int radius)
{
	// Plot a circle using Bresenham's algorithm
	// Note: No trigonometry, floating point numbers or division used
	// Note: Circle is clipped against top, bottom, left and right edge.
	int PlotCircle_rval;
	int PlotCircle_x1;
	int PlotCircle_y1;
	__asm {
		// Input: p1x, p1y, radius, color, buffer, height, width
		// Only 1/8 of the full circle is calculated, each pixel is mirrored 7 times
		// to produce a perfect circle at the shortest possible time
		// Originally translated from Pascal to 16bit asm by TheJoker/ATC in 1995
		// Rewritten for 32bit assembly by Andy Floyd
		mov eax,radius
		cmp eax,0			// Zero or negative radius?
		jle Skip			// Yes, this is an invalid circle

		shl eax,1
		sub eax,3
		neg eax
		mov PlotCircle_rval,eax
		mov PlotCircle_x1,0
		mov eax,radius
		mov PlotCircle_y1,eax

		// Section 1 (SSE)
			PlotCircle_loop:	
		mov eax,PlotCircle_y1
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotCircle_not2	
		mul width

		mov edx,PlotCircle_x1
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jae PlotCircle_not1	
		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 2 (SSW)
			PlotCircle_not1:	
		mov eax,PlotCircle_y1
		add eax,p1y
		mul width

		mov edx,PlotCircle_x1
		neg edx
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jae PlotCircle_not2	
		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 3 (NNW)
			PlotCircle_not2:
		mov eax,PlotCircle_y1
		neg eax
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotCircle_not4	
		mul width

		mov edx,PlotCircle_x1
		neg edx
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jae PlotCircle_not3	
		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 4 (NNE)
			PlotCircle_not3:
		mov eax,PlotCircle_y1
		neg eax
		add eax,p1y
		mul width

		mov edx,PlotCircle_x1
		add edx,p1x
		add eax,edx
		cmp edx,width
		jae PlotCircle_not4	

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 5 (ESE)
			PlotCircle_not4:
		mov eax,PlotCircle_x1
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotCircle_not6	
		mul width

		mov edx,PlotCircle_y1
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jae PlotCircle_not5	

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 6 (WSW)
			PlotCircle_not5:
		mov eax,PlotCircle_x1
		add eax,p1y
		mul width

		mov edx,PlotCircle_y1
		neg edx
		add edx,p1x
		add eax,edx
		cmp edx,width		// Outside right edge?
		jae PlotCircle_not6	

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 7 (WNW)
			PlotCircle_not6:
		mov eax,PlotCircle_x1
		neg eax
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotCircle_not8	
		mul width

		mov edx,PlotCircle_y1
		neg edx
		add edx,p1x
		add eax,edx
		cmp edx,width		// Outside right edge?
		jae PlotCircle_not7	

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 8 (ENE)
			PlotCircle_not7:
		mov eax,PlotCircle_x1
		neg eax
		add eax,p1y
		mul width

		mov edx,PlotCircle_y1
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jae PlotCircle_not8	

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

			PlotCircle_not8:

		// Calculate next pixel coordinate
		cmp PlotCircle_rval,0
		jnl PlotCircle_skip1
		mov eax,PlotCircle_x1
		shl eax,2
		add eax,6
		add PlotCircle_rval,eax
		jmp PlotCircle_skip2
			PlotCircle_skip1:
		mov eax,PlotCircle_x1
		sub eax,PlotCircle_y1
		shl eax,2
		add eax,10
		add PlotCircle_rval,eax
		dec PlotCircle_y1
			PlotCircle_skip2:
		inc PlotCircle_x1
		mov eax,PlotCircle_x1
		cmp eax,PlotCircle_y1
		jna PlotCircle_loop	

		Skip:	// Jump here if radius is zero
	}
}	// plot_circle_xy()




void fill_circle(int p1x, int p1y, int radius)
{
	// Plot a solid filled circle using Bresenham's algorithm
	// Note: No trigonometry, floating point numbers or division used
	// Note: This function does NOT use the PolyBuffer
	// Note: Circle is clipped against top, bottom, left and right edge.
	int PlotDisc_rval;
	int PlotDisc_x1;
	int PlotDisc_y1;
	__asm {

		// Input: p1x, p1y, radius, color, buffer, height, width
		// Only 1/8 of the full circle is calculated, each point is mirrored 7 times
		// and finally a horizontal line drawn between them
		mov eax,radius
		cmp eax,0		// Zero or negative radius?
		jle Skip		// Yes, this is an invalid circle

		shl eax,1
		sub eax,3
		neg eax
		mov PlotDisc_rval,eax
		mov PlotDisc_x1,0
		mov eax,radius
		mov PlotDisc_y1,eax

		// Section 1 (SSE)
			PlotDisc_loop:	
		mov eax,PlotDisc_y1
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotDisc_not2	
		mul width

		mov edx,PlotDisc_x1
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jb  PlotDisc_mark1
		mov ecx,width
		jmp Plotdisc_not1
			PlotDisc_mark1:
		mov ecx,edx		// East end of line marked

		// Section 2 (SSW)
			PlotDisc_not1:	
		mov eax,PlotDisc_y1
		add eax,p1y
		mul width

		mov edx,PlotDisc_x1
		neg edx
		add edx,p1x
		mov ebx,eax
		add eax,edx
		cmp edx,width		// Outside left edge?
		jb  PlotDisc_mark2
		mov eax,ebx
		jmp PlotDisc_plot2
			PlotDisc_mark2:
		sub ecx,edx
			PlotDisc_plot2:

		mov edi,buffer	// Now draw a line from eax, ecx pixels long
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		rep stosd

		// Section 3 (NNW)
			PlotDisc_not2:
		mov eax,PlotDisc_y1
		neg eax
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotDisc_not4	
		mul width

		mov edx,PlotDisc_x1
		neg edx
		add edx,p1x
		mov ebx,eax
		add eax,edx
		cmp edx,width	// Outside right edge?
		jb  PlotDisc_mark3
		mov eax,ebx
			PlotDisc_mark3:
		mov ecx,eax

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax

		// Section 4 (NNE)
		mov eax,PlotDisc_y1
		neg eax
		add eax,p1y
		mul width

		mov edx,PlotDisc_x1
		add edx,p1x
		mov ebx,eax
		cmp edx,width	// Outside right edge?
		jbe PlotDisc_mark4	// Within boundaries

		add eax,width
		jmp PlotDisc_plot4
			PlotDisc_mark4:
		add eax,edx
			PlotDisc_plot4:
				
		mov ebx,eax		// ebx = eax(sluttpunkt)
		sub ebx,ecx		// ebx = ebx(sluttpunkt) - ecx(estartpunkt)
		mov eax,ecx		// eax = ecx(startpunkt)
		mov ecx,ebx		// ecx = ebx(lengde)

		mov edi,buffer
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax
		rep stosd

		// Section 5 (ESE)
			PlotDisc_not4:
		mov eax,PlotDisc_x1
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotDisc_not6	
		mul width

		mov edx,PlotDisc_y1
		add edx,p1x
		add eax,edx
		cmp edx,width	// Outside right edge?
		jb  PlotDisc_mark5
		mov ecx,width
		jmp Plotdisc_not5
			PlotDisc_mark5:
		mov ecx,edx		// East end of line marked

		// Section 6 (WSW)
			PlotDisc_not5:
		mov eax,PlotDisc_x1
		add eax,p1y
		mul width

		mov edx,PlotDisc_y1
		neg edx
		add edx,p1x
		mov ebx,eax
		add eax,edx
		cmp edx,width		// Outside left edge?
		jb  PlotDisc_mark6
		mov eax,ebx
		jmp PlotDisc_plot6
			PlotDisc_mark6:
		sub ecx,edx
			PlotDisc_plot6:

		mov edi,buffer		// Now draw a line from eax, ecx pixels long
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		rep stosd

		// Section 7 (WNW)
			PlotDisc_not6:
		mov eax,PlotDisc_x1
		neg eax
		add eax,p1y
		cmp eax,height	// Below bottom edge?
		jae PlotDisc_not8	
		mul width

		mov edx,PlotDisc_y1
		neg edx
		add edx,p1x
		mov ebx,eax
		add eax,edx
		cmp edx,width		// Outside left edge?
		jb  PlotDisc_mark7
		mov eax,ebx
			PlotDisc_mark7:
		mov ecx,eax

		// Section 8 (ENE)
		mov eax,PlotDisc_x1
		neg eax
		add eax,p1y
		mul width

		mov edx,PlotDisc_y1
		add edx,p1x
		mov ebx,eax
		cmp edx,width	// Outside right edge?
		jbe PlotDisc_mark8	// Within boundaries

		add eax,width
		jmp PlotDisc_plot8
			PlotDisc_mark8:
		add eax,edx
			PlotDisc_plot8:
				
		mov ebx,eax		// ebx = eax(sluttpunkt)
		sub ebx,ecx		// ebx = ebx(sluttpunkt) - ecx(estartpunkt)
		mov eax,ecx		// eax = ecx(startpunkt)
		mov ecx,ebx		// ecx = ebx(lengde)

		mov edi,buffer	// Plot a line from eax, ecx pixels long
		shl eax,2	// 4 bytes per pixel
		add edi,eax
		mov eax,color
		mov [edi],eax
		rep stosd

			PlotDisc_not8:

		// Calculate next pixel coordinate
		cmp PlotDisc_rval,0
		jnl PlotDisc_skip1
		mov eax,PlotDisc_x1
		shl eax,2
		add eax,6
		add PlotDisc_rval,eax
		jmp PlotDisc_skip2
			PlotDisc_skip1:
		mov eax,PlotDisc_x1
		sub eax,PlotDisc_y1
		shl eax,2
		add eax,10
		add PlotDisc_rval,eax
		dec PlotDisc_y1
			PlotDisc_skip2:
		inc PlotDisc_x1
		mov eax,PlotDisc_x1
		cmp eax,PlotDisc_y1
		jna PlotDisc_loop	

		Skip:	// Jump here if radius is zero
	}
}	// plot_disc_xy



void fill_rectangle(int p1x, int p1y, int p2x, int p2y)
{
	// Plot a filled rectangle after flipping/clipping as necessary
	// Input: p1x, p1y, p2x, p2y, color, buffer, height, width
	// Note: This function does NOT use the PolyBuffer
	// Note: Rectangle is clipped against top, bottom, left and right edge.
	__asm {
		mov eax,p1x
		cmp eax,p2x			// Compare p1x and p2x
		jc  PlotBlock_skip1
		mov eax,p1x
		mov ebx,p2x
		mov p1x,ebx
		mov p2x,eax
			PlotBlock_skip1:	// p2x is now >= p1x
		mov eax,p1y
		cmp eax,p2y			// Compare p1y and p2y
		jc  PlotBlock_skip2
		mov eax,p1y
		mov ebx,p2y
		mov p1y,ebx
		mov p2y,eax
			PlotBlock_skip2:	// p2y us now >= p1y

		// Trivial check; skip rectangles that lie outside the viewing area
		mov eax,p1y
		cmp eax,height	// Entire rectangle below bottom edge?
		jge PlotBlock_skip

		mov eax,p2y
		cmp eax,0		// Entire rectangle above top edge?
		jl PlotBlock_skip
		
		mov eax,p1x
		cmp eax,width	// Entire rectangle outside right edge?
		jge PlotBlock_skip

		mov eax,p2x
		cmp eax,0		// Entire rectangle outside left edge?
		jl PlotBlock_skip

		// Clip rectangles that lie partially outside the viewing area
		mov eax,p1y
		cmp eax,0
		jge PlotBlock_topok
		xor eax,eax		// eax = 0
		mov p1y,eax
			PlotBlock_topok:

		mov eax,p2y
		cmp eax,height
		jl PlotBlock_bottomok
		mov eax,height
		mov p2y,eax
			PlotBlock_bottomok:

		mov eax,p1x
		cmp eax,0
		jge PlotBlock_leftok
		xor eax,eax		// eax = 0
		mov p1x,eax
			PlotBlock_leftok:

		mov eax,p2x
		cmp eax,width
		jl PlotBlock_rightok
		mov eax,width
		mov p2x,eax
			PlotBlock_rightok:

		// Do the actual plotting
		mov ebx,p2x
		sub ebx,p1x
		inc ebx			// Zero width means plot one pixel

		mov ecx,p2y
		sub ecx,p1y
		inc ecx			// Zero height means plot one line
				
		mov edi,buffer
		mov eax,p1y
		mul width
		add eax,p1x
		shl eax,2		// Each pixel is 4 bytes
		add edi,eax

		mov edx,256
		mov edx,width
		sub edx,ebx
		shl edx,2		// Each pixel is 4 bytes
		// edi now points at upper left corner, edx holds 'newline' offset
		// ecx holds the number of lines to plot and ebx holds the length of each line

			PlotBlock_loop1:
		push ecx		// Push remaining height to stack
		mov ecx,ebx		// eax = width
		mov eax,color	
		rep stosd		// Plot one line
		add edi,edx		// Skip to beginning of next line
		pop ecx			// ecx = Number of lines left to plot
		loop PlotBlock_loop1

			PlotBlock_skip:
	}
}	// plot_block()


void plot_line(int p1x, int p1y, int p2x, int p2y)
{
	// Plot a single straight line using Bresenham's algorithm
	// Input: p1x, p1y, p2x, p2y, color, buffer, width
	// Note: No trigonometry, floating point numbers or division used
	// Note: Line is clipped against top, bottom, left and right edge.
	int PlotLine_LgDelta;
	int PlotLine_LgStep;
	int PlotLine_ShDelta;
	int PlotLine_ShStep;
	int PlotLine_Cycle;
	__asm {
		mov eax,p1x	// Line must be plotted from left to right
		cmp p2x,eax
	    jg  PlotLine_noswap
		mov edx,p1x
		mov eax,p2x
		mov p1x,eax
		mov p2x,edx
		mov edx,p1y
		mov eax,p2y
		mov p1y,eax
		mov p2y,edx
			PlotLine_noswap:

		mov eax,p2x
		sub eax,p1x
		mov PlotLine_LgDelta,eax	// LgDelta = total line width

		mov eax,p2y
		sub eax,p1y
		mov PlotLine_ShDelta,eax	// ShDelta = total line height

		cmp PlotLine_LgDelta,0
		jge PlotLine_skip1
		neg PlotLine_LgDelta
		mov PlotLine_LgStep,-1
		jmp PlotLine_skip2
			PlotLine_skip1:
		mov PlotLine_LgStep,1
			PlotLine_skip2:

		cmp PlotLine_ShDelta,0
		jge PlotLine_skip3
		neg PlotLine_ShDelta
		mov PlotLine_ShStep,-1
		jmp PlotLine_skip4
			PlotLine_skip3:
		mov PlotLine_ShStep,1
			PlotLine_skip4:

		// Choose method 1 or 2 depending on the line slope
		mov eax,PlotLine_ShDelta
		cmp PlotLine_LgDelta,eax
		jbe PlotLine_skip5		// Increment along Y axis

		// Method 1: Increment along X axis
		mov eax,PlotLine_LgDelta
		shr eax,1
		mov PlotLine_Cycle,eax

			PlotLine_loop1:
		// Plot this point
		mov ebx,p1x
		cmp ebx,0
		jl PlotLine_clipped1	// outside left edge
		cmp ebx,width
		jae PlotLine_clipped1	// outside right edge
		mov ecx,p1y
		cmp ecx,0
		jl PlotLine_clipped1	// outside top edge
		cmp ecx,height
		jae PlotLine_clipped1	// outside bottom edge

		mov eax,width	// eax = width
		mul ecx			// edx:eax = width * ecx
		add eax,ebx		// eax = eax + ebx
		shl eax,2		// 4 bytes per pixel
		mov edi,buffer
		add edi,eax
		mov eax,color
		mov [edi],eax
			PlotLine_clipped1:

		mov eax,p1x				// Have we reached the endpoint yet?
		cmp eax,p2x
		je  PlotLine_finished	// Yes

		mov eax,PlotLine_LgStep
		add p1x,eax
		mov eax,PlotLine_ShDelta
		add PlotLine_Cycle,eax

		mov eax,PlotLine_LgDelta
		cmp PlotLine_Cycle,eax
		jl  PlotLine_skip7
		mov eax,PlotLine_ShStep
        add p1y,eax

		mov eax,PlotLine_LgDelta
		sub PlotLine_Cycle,eax
			PlotLine_skip7:
		jmp PlotLine_loop1

		// ----
			PlotLine_skip5:
		// Method 2: Increment along Y axis
		mov eax,PlotLine_ShDelta
		shr eax,1
		mov PlotLine_Cycle,eax
		mov eax,PlotLine_LgDelta
		mov ebx,PlotLine_ShDelta
		mov PlotLine_LgDelta,ebx
		mov PlotLine_ShDelta,eax

		mov eax,PlotLine_LgStep
		mov ebx,PlotLine_ShStep
		mov PlotLine_LgStep,ebx
		mov PlotLine_ShStep,eax

			PlotLine_loop2:
		// Plot this point
		mov ebx,p1x
		cmp ebx,0
		jl PlotLine_clipped2	// outside left edge
		cmp ebx,width
		jae PlotLine_clipped2	// outside right edge
		mov ecx,p1y
		cmp ecx,0
		jl PlotLine_clipped2	// outside top edge
		cmp ecx,height
		jae PlotLine_clipped2	// outside bottom edge

		mov ebx,p1x
		mov ecx,p1y
		mov eax,width	// eax = width
		mul ecx			// edx:eax = width * ecx
		add eax,ebx		// eax = eax + ebx
		shl eax,2		// 4 bytes per pixel
		mov edi,buffer
		add edi,eax
		mov eax,color
		mov [edi],eax
			PlotLine_clipped2:

		mov eax,p2y				// Have we reached the endpoint yet?
		cmp p1y,eax
		je  PlotLine_finished	// Yes

		mov eax,PlotLine_LgStep
		add p1y,eax
		mov eax,PlotLine_ShDelta
		add PlotLine_Cycle,eax

		mov eax,PlotLine_LgDelta
		cmp PlotLine_Cycle,eax
		jl  PlotLine_skip9
		mov eax,PlotLine_ShStep
		add p1x,eax

		mov eax,PlotLine_LgDelta
		sub PlotLine_Cycle,eax
			PlotLine_skip9:
		jmp PlotLine_loop2

			PlotLine_finished:

	}
}	//plot_line()


void plot_rectangle(int p1x, int p1y, int p2x, int p2y)
{
	// Plot a rectangle after flipping/clipping as necessary
	// Input: p1x, p1y, p2x, p2y, color, buffer, height, width
	// Note: Rectangle is clipped against top, bottom, left and right edge.
	plot_line(p1x,p1y,p2x,p1y);	// top
	plot_line(p1x,p2y,p2x,p2y);	// bottom
	plot_line(p1x,p1y,p1x,p2y);	// left
	plot_line(p2x,p1y,p2x,p2y);	// right
}	// plot_rectangle()



void plot_triangle(int p1x, int p1y, int p2x, int p2y, int p3x, int p3y)
{
	// Plot a triangle using three calls to plot_line()
	// Note: Triangle is clipped against top, bottom, left and right edge.
	plot_line(p1x, p1y, p2x, p2y);
	plot_line(p2x, p2y, p3x, p3y);
	plot_line(p3x, p3y, p1x, p1y);
}	// plot_triangle()


void reset_polybuffer()
{
	// Reset the rasterline buffer used to fill polygons
	// This MUST be done before starting to plot each polygon
	// Note: PolyBufferHi and PolyBufferLo are used to determine how much
	// of the buffer has been used. To force a full buffer reset, set those
	// to MAX_SCANLINES-1 and 0 respectively. This should never be necessary.
	unsigned int scanline;
	// Do bounds checking in case we screwed up anywhere
	if (PolyBufferHi > MAX_SCANLINES - 1) { PolyBufferHi = MAX_SCANLINES - 1; }
	if (PolyBufferLo < 0) { PolyBufferLo =  0; }
	for (scanline=PolyBufferLo; scanline<PolyBufferHi; scanline++)
	{
		PolyBuffer[scanline*2] = -1;
		PolyBuffer[(scanline*2)+1] = -1;
	}

	PolyBufferHi = 0;
	PolyBufferLo = MAX_SCANLINES - 1;

}	// clear_polybuffer


void plot_polybuffer()
{
	// Plot raster lines PolyBufferLo .. PloyBufferHi
	// as described by PolyBuffer to produce a filled polygon
	// (or any other concave shape for that matter)
	// Note: Plotted area is clipped against left and right edge.
	// Warning: Lo must NEVER be greater than Hi
	int unsigned scanline;
	int unsigned from;
	int unsigned to;

	// Do additional bounds checking in case we screwed up anywhere
	PolyBufferHi++;
	if (PolyBufferHi > MAX_SCANLINES - 1) { PolyBufferHi = MAX_SCANLINES - 1; }
	if (PolyBufferLo < 0) { PolyBufferLo =  0; }

	if (PolyBufferHi > PolyBufferLo) 
	{

		// Process each scan line
		for (scanline=PolyBufferLo; scanline<PolyBufferHi; scanline++)
		{
			from = PolyBuffer[scanline*2];
			to = PolyBuffer[(scanline*2)+1];

			if (from != -1 && to != -1)	// Scan line valid?
			{
				if (from == -2) { from = 0; }			// Clipped to the left?
				if (to == width) { to = width - 1; }	// Clipped to the right?
				if (to >= from) 
				{

					// Plot this scan line
					__asm{
						mov edi,buffer
						mov eax,scanline	// y
						mul width
						add eax,from		// x
						shl eax,2		// 4 bytes per pixel
						add edi,eax		// edi now points to start of scan line

						mov ecx,to
						sub ecx,from
						inc ecx

						mov eax,color
						rep stosd
					}
				}
			}
		}
	}
	
}


void plot_leg(int p1x, int p1y, int p2x, int p2y)
{
	// Plot a single straight line into the PolyBuffer using Bresenham's algorithm
	// Input: p1x, p1y, p2x, p2y, buffer, width
	// Note: No trigonometry, floating point numbers or division used
	// Note: Leg is clipped against top and bottom.
	int FillPoly_LgDelta;
	int FillPoly_LgStep;
	int FillPoly_ShDelta;
	int FillPoly_ShStep;
	int FillPoly_Cycle;
	__asm {
		mov eax,p1x	// Line must be plotted from left to right
		cmp p2x,eax
	    jg  FillPoly_noswap
		mov edx,p1x
		mov eax,p2x
		mov p1x,eax
		mov p2x,edx
		mov edx,p1y
		mov eax,p2y
		mov p1y,eax
		mov p2y,edx
			FillPoly_noswap:

		mov eax,p2x
		sub eax,p1x
		mov FillPoly_LgDelta,eax	// LgDelta = total line width

		mov eax,p2y
		sub eax,p1y
		mov FillPoly_ShDelta,eax	// ShDelta = total line height

		cmp FillPoly_LgDelta,0
		jge FillPoly_skip1
		neg FillPoly_LgDelta
		mov FillPoly_LgStep,-1
		jmp FillPoly_skip2
			FillPoly_skip1:
		mov FillPoly_LgStep,1//1
			FillPoly_skip2:

		cmp FillPoly_ShDelta,0
		jge FillPoly_skip3
		neg FillPoly_ShDelta
		mov FillPoly_ShStep,-1
		jmp FillPoly_skip4
			FillPoly_skip3:
		mov FillPoly_ShStep,1
			FillPoly_skip4:

		mov eax,FillPoly_ShDelta
		cmp FillPoly_LgDelta,eax
		jle FillPoly_skip5		// Increment along Y axis

		// Increment along X axis
		mov eax,FillPoly_LgDelta
		shr eax,1
		mov FillPoly_Cycle,eax

			FillPoly_loop1:
		mov ecx,p1y
		cmp ecx,0
		jl FillPoly_clipped1	// outside top edge
		cmp ecx,height
		jge FillPoly_clipped1	// outside bottom edge

		call FillPolyPoint		// Mark this point in PolyBuffer
			FillPoly_clipped1:

		mov eax,p1x				// Have we reached the endpoint yet?
		cmp eax,p2x
		je  FillPoly_finished	// Yes

			mov eax,FillPoly_LgStep
		add p1x,eax
		mov eax,FillPoly_ShDelta
		add FillPoly_Cycle,eax

		mov eax,FillPoly_LgDelta
		cmp FillPoly_Cycle,eax
		jl  FillPoly_skip7		// More horizontal pixels before vertical?

		mov eax,FillPoly_ShStep
        add p1y,eax

		mov eax,FillPoly_LgDelta
		sub FillPoly_Cycle,eax
			FillPoly_skip7:
		jmp FillPoly_loop1

		// ----
			FillPoly_skip5:
		// Increment along Y axis
		mov eax,FillPoly_ShDelta
		shr eax,1
		mov FillPoly_Cycle,eax
		mov eax,FillPoly_LgDelta
		mov ebx,FillPoly_ShDelta
		mov FillPoly_LgDelta,ebx
		mov FillPoly_ShDelta,eax

		mov eax,FillPoly_LgStep
		mov ebx,FillPoly_ShStep
		mov FillPoly_LgStep,ebx
		mov FillPoly_ShStep,eax

			FillPoly_loop2:

		mov ecx,p1y
		cmp ecx,0
		jl  FillPoly_clipped2	// outside top edge
		cmp ecx,height
		jge FillPoly_clipped2	// outside bottom edge

		call FillPolyPoint		// Mark this point in PolyBuffer
			FillPoly_clipped2:


		mov eax,p2y				// Have we reached the endpoint yet?
		cmp p1y,eax
		je  FillPoly_finished	// Yes

		mov eax,FillPoly_LgStep
		add p1y,eax
		mov eax,FillPoly_ShDelta
		add FillPoly_Cycle,eax

		mov eax,FillPoly_LgDelta
		cmp FillPoly_Cycle,eax
		jl  FillPoly_skip9
		mov eax,FillPoly_ShStep
		add p1x,eax

		mov eax,FillPoly_LgDelta
		sub FillPoly_Cycle,eax
			FillPoly_skip9:
		jmp FillPoly_loop2


			FillPolyPoint:		// --- sub routine ---
		// Check PolyBuffer. There are two values here; lowest/highest x value
		mov edi,offset PolyBuffer
		mov eax,p1y		// current scanline
		shl eax,3		// 2 dwords per scanline
		add edi,eax		// edi now points at first value
		mov ebx,p1x		// ebx = x
		// Does this point clip outside the left edge?
		cmp ebx,0
		jg  PolyPoint_insideleft
		mov ebx,-2
		mov [edi],ebx	// Yes, use -2 as the start value and exit
		jmp PolyPoint_EndOk
			PolyPoint_insideleft:

		// Does this point clip outside the right edge?
		mov ecx,width
		cmp ebx,ecx
		jle	PolyPoint_insideright
		mov ebx,width
		mov [edi+4],ebx	// Yes, use width as the end value and exit
		jmp PolyPoint_EndOk

			PolyPoint_insideright:
		// The point is within the valid area
		mov ecx,[edi]	// ecx = start
		mov edx,[edi+4]	// edx = end

		// edi now points to the correct scanline buffer
		// ecx and edx holds the start and stop points respectively
		// ebx holds the new x coordinate

		// Examine start value
		cmp ecx,-1					// Do we have a start value already?
		jne  PolyPoint_StartExists
		mov [edi],ebx				// No, use new point
		jmp PolyPoint_StartOk
			PolyPoint_StartExists:

		cmp ecx,ebx
		jle  PolyPoint_StartOk
		mov [edi],ebx

			PolyPoint_StartOk:
		// Examine end value
		cmp edx,-1					// Do we have an end value already?
		jne  PolyPoint_EndExists
		mov [edi+4],ebx				// No, use new point
		jmp PolyPoint_EndOk
			PolyPoint_EndExists:

		cmp edx,ebx
		jge  PolyPoint_EndOk
		mov [edi+4],ebx

			PolyPoint_EndOk:

		// Update Hi and Lo values depending on p1y
		mov eax,p1y
		cmp eax,PolyBufferHi
		jl  P1y_is_greater
		mov PolyBufferHi,eax
			P1y_is_greater:
		cmp eax,PolyBufferLo
		jg	P1y_is_lower
		mov PolyBufferLo,eax
			P1y_is_lower:					
		ret

			FillPoly_finished:
	}
}	// plot_leg()


void fill_triangle(int p1x, int p1y, int p2x, int p2y, int p3x, int p3y)
{
	// Plot a filled triangle using three calls to plot_leg()
	// Note: This function uses the PolyBuffer
	// Note: Triangle is clipped against top, bottom, left and right edge.
	reset_polybuffer();
	plot_leg(p1x, p1y, p2x, p2y);
	plot_leg(p2x, p2y, p3x, p3y);
	plot_leg(p3x, p3y, p1x, p1y);
	plot_polybuffer();
}	// fill_triangle()


void plot_polygon(int points, int *pNx, int *pNy)
{
	// Plot a convex polygon with 3 or more points/sides using color
	// Note: Polygon is clipped against top, bottom, left and right edge.
	int index;
	for (index=1; index<points; index++)
	{
		plot_line(pNx[index-1], pNy[index-1], pNx[index], pNy[index]);	// Connect the dots
	}
	plot_line(pNx[points-1], pNy[points-1], pNx[0], pNy[0]);			// Complete the loop
}


void fill_polygon(int points, int *pNx, int *pNy)
{
	// Plot a CONVEX polygon with 3 or more points/sides using color
	// Note: This function uses the PolyBuffer
	// Note: Polygon is clipped against top, bottom, left and right edge.
	// WARNING: concave polygons WILL BE PLOTTED INCORRECTLY
	int index;
	reset_polybuffer();
	for (index=1; index<points; index++)
	{
		plot_leg(pNx[index-1], pNy[index-1], pNx[index], pNy[index]);	// Connect the dots
	}
	plot_leg(pNx[points-1], pNy[points-1], pNx[0], pNy[0]);			// Complete the loop
	plot_polybuffer();
}


void unload_font()
{
	// Deallocate font buffers if in use
	// Note: This is safe to use even if no font has been loaded
	// so it's always successful
	if (font_bitmap != NULL) { free(font_bitmap); }
	if (font_index  != NULL) { free(font_index);  }
	if (font_size   != NULL) { free(font_size);   }
	font_bitmap_length = 0;
	font_index_length = 0;
	font_size_length = 0;
}



char load_font(char *font)
{
	// Load a named font into memory
	// Note: Loading a font automatically unloads any previously loaded font.
	// Returns 1 if the operation was successful, 0 otherwise
	FILE *filehandle;
	char filename[260];
//	char *path = "release\\resource\\fonts\\";
	char *path = "";

	char loaded = 0;	// Number of files loaded OK
	int filesize;		// Used for determining file/buffer size

	// Before we load any new files, make sure any existing font is deallocated
	unload_font();

	// Each font is made up from 3 files:
	// .fon is the bitmap
	// .idx maps each ASCII code to a bitmap offset
	// .siz gives the bitmap width of each ASCII char
	// See resource/fonts/format.txt for more details

	// Load font BITMAP file
	filename[0] = 0;			// filename = ""
	strcat(filename, path);		// filename .= path
	strcat(filename, font);		// filename .= fontname
	strcat(filename, ".fon");	// filename .= .fon
	filehandle = fopen(filename, "rb");
	if (filehandle != NULL)
	{
		// Determine the file size
		fseek(filehandle, 0, SEEK_END);		// Seek to the end...
		filesize = ftell(filehandle);		// ...to find out how big the file is
		rewind(filehandle);					// Rewind so we can read the file
		if (filesize > 0) 
		{
			// Allocate a buffer
			font_bitmap = (unsigned char *) malloc(filesize);
			if (font_bitmap != NULL) 
			{
				// Read the .fon file into font_bitmap
				font_bitmap_length = fread(font_bitmap, 1, filesize, filehandle);
				loaded++;
			}
		}
		fclose(filehandle);
	}

	// Load font INDEX file
	filename[0] = 0;			// filename = ""
	strcat(filename, path);		// filename .= path
	strcat(filename, font);		// filename .= fontname
	strcat(filename, ".idx");	// filename .= .idx
	filehandle = fopen(filename, "rb");
	if (filehandle != NULL)
	{
		// Determine the file size
		fseek(filehandle, 0, SEEK_END);
		filesize = ftell(filehandle);
		rewind(filehandle);
		if (filesize > 0) 
		{
			// Allocate a buffer
			font_index = (unsigned char *) malloc(filesize);
			if (font_index != NULL) 
			{
				// Read the .idx file into font_index
				font_index_length = fread(font_index, 1, filesize, filehandle);
				loaded++;
			}
		}
		fclose(filehandle);
	}

	// Load font SIZE file
	filename[0] = 0;			// filename = ""
	strcat(filename, path);		// filename .= path
	strcat(filename, font);		// filename .= fontname
	strcat(filename, ".siz");	// filename .= .siz
	filehandle = fopen(filename, "rb");
	if (filehandle != NULL)
	{
		// Determine the file size
		fseek(filehandle, 0, SEEK_END);
		filesize = ftell(filehandle);
		rewind(filehandle);
		if (filesize > 0) 
		{
			// Allocate a buffer
			font_size = (unsigned char *) malloc(filesize);
			if (font_size != NULL) 
			{
				// Read the .siz file into font_size
				font_size_length = fread(font_size, 1, filesize, filehandle);
				loaded++;
			}
		}
		fclose(filehandle);
	}

	// If we failed, back off and free all the buffers
	if (loaded < 3) { 
		unload_font(); 
		return 0;	// Failed
	}	
	else
	{
		return 1;	// OK
	}
}


void plot_char(unsigned char ascii)
{
	// Plot a single ASCII character using the current width, font and color
	// The position is determined by cursor_x and cursor_y
	// Note: Chars are clipped against the top, bottom, left and right edge
	// Currently, "half characters" can only occur at the right edge
	unsigned short int start;
	unsigned char startlo;
	unsigned char starthi;
	unsigned short int index;
	unsigned char size;
	unsigned char bitmap;	// Each byte in font_bitmap describes 8 vertical pixels

	// Trivial clipping; The char starts off the left or top edge
	if (cursor_x < 0 || cursor_y < 0) {
		return;	// Clipped, do nothing
	}

	// Trivial clipping; The char starts off the right or bottom edge
	if ((cursor_x >= (int) width) || (cursor_y >= (int) (height-8))) {
		return;	// Clipped, do nothing
	}

	if (font_bitmap != NULL && font_index != NULL && font_size != NULL) {

		startlo = (unsigned char) *(font_index + (ascii * 2));
		starthi = (unsigned char) *(font_index + (ascii * 2) + 1);
		start = (unsigned short int) (starthi * 256) + startlo;
		size = (unsigned char) *(font_size + ascii);

		// Plot each vertical "slice" of this char
		for (index=0; index<size; index++)
		{
			bitmap = (unsigned char) *(font_bitmap + start + index);
			__asm {
				mov edi,buffer
				mov eax,cursor_y
				mul width
				add eax,cursor_x
				shl eax,2
				add edi,eax			// edi points to cursor_x, cursor_y

				mov ecx, 8			// we will examine all 8 bits in "stripe"

				mov dl,bitmap
				mov eax,color		// eax = color
				mov ebx,width
				shl ebx,2			// ebx = width (4 bytes per pixel)
	
					Char_loop:
				mov dh,dl
				and dh,1
				cmp dh,0
				je Skip_pixel
				mov [edi],eax

					Skip_pixel:
				add edi,ebx
				shr dl,1

				loop Char_loop;

			}
			cursor_x = cursor_x + 1;
			if (cursor_x == (int) width) { return; }	// Abort here if we've reached the buffer edge
		}

	}
	else
	{
		// Write "no font" using lines.... the only way we can communicate :-/
		plot_line(cursor_x, cursor_y, cursor_x, cursor_y+8);
		plot_line(cursor_x, cursor_y, cursor_x+8, cursor_y+8);
		plot_line(cursor_x+8, cursor_y, cursor_x+8, cursor_y+8);

		plot_line(cursor_x+10, cursor_y, cursor_x+10, cursor_y+8);
		plot_line(cursor_x+10, cursor_y, cursor_x+18, cursor_y);
		plot_line(cursor_x+18, cursor_y, cursor_x+18, cursor_y+8);
		plot_line(cursor_x+10, cursor_y+8, cursor_x+18, cursor_y+8);

		plot_line(cursor_x+30, cursor_y, cursor_x+30, cursor_y+8);
		plot_line(cursor_x+30, cursor_y, cursor_x+38, cursor_y);
		plot_line(cursor_x+30, cursor_y+4, cursor_x+38, cursor_y+4);

		plot_line(cursor_x+40, cursor_y, cursor_x+40, cursor_y+8);
		plot_line(cursor_x+40, cursor_y, cursor_x+48, cursor_y);
		plot_line(cursor_x+48, cursor_y, cursor_x+48, cursor_y+8);
		plot_line(cursor_x+40, cursor_y+8, cursor_x+48, cursor_y+8);

		plot_line(cursor_x+50, cursor_y, cursor_x+50, cursor_y+8);
		plot_line(cursor_x+50, cursor_y, cursor_x+58, cursor_y+8);
		plot_line(cursor_x+58, cursor_y, cursor_x+58, cursor_y+8);

		plot_line(cursor_x+60, cursor_y, cursor_x+68, cursor_y);
		plot_line(cursor_x+64, cursor_y, cursor_x+64, cursor_y+8);

		cursor_x = cursor_x + 80;
	}


	// Handle \t
	if (ascii == 9)
	{
		cursor_x = ((cursor_x + 80) / 80) * 80;	// Move to next tab stop
	}

	// Handle \n
	if (ascii == 10) 
	{
		cursor_x = 1;							// Move to beginning of next line
		cursor_y = cursor_y + 10;
	}

}


void plot_string(unsigned char *string, int p1x, int p1y)
{
	// Use plot_char to plot the string at the chosen coordinates
	// If p1x and/or p1y is -1 then use cursor_x/cursor_y instead
	// Note: Chars are clipped against the top, bottom, left and right edge
	// Currently, "half characters" can only occur at the right edge
	int index;

	if (p1x != -1) { cursor_x = p1x; }
	if (p1y != -1) { cursor_y = p1y; }

	for (index=0; index<MAX_STRING_LENGTH; index++)
	{
		if (string[index] == 0) { break; }	// Reached string terminator
		plot_char(string[index]);
	}
}



void texture_polybuffer(int texture)
{
	// Plot raster lines PolyBufferLo .. PloyBufferHi
	// as described by PolyBuffer to produce a textured polygon
	// (or any other concave shape for that matter)
	// Note: Plotted area is clipped against left and right edge.
	// Warning: Lo must NEVER be greater than Hi
	int unsigned scanline;
	int unsigned from;
	int unsigned to;

	char debug[256];

	// Do additional bounds checking in case we screwed up anywhere
	PolyBufferHi++;
	if (PolyBufferHi > MAX_SCANLINES - 1) { PolyBufferHi = MAX_SCANLINES - 1; }
	if (PolyBufferLo < 0) { PolyBufferLo =  0; }

	if (PolyBufferHi > PolyBufferLo) 
	{
		// DEBUG: Show texture number we're working with		
		sprintf(debug, "Texture number %i", texture);
		color = WHITE;
		plot_string(debug, 1,10 + (10*texture));

		// Process each scan line
		for (scanline=PolyBufferLo; scanline<PolyBufferHi; scanline++)
		{
			from = PolyBuffer[scanline*2];
			to = PolyBuffer[(scanline*2)+1];

			if (from != -1 && to != -1)	// Scan line valid?
			{
				if (from == -2) { from = 0; }			// Clipped to the left?
				if (to == width) { to = width - 1; }	// Clipped to the right?
				if (to >= from) 
				{

					// Plot texture data

				}
			}
		}
	}
	
}




void fill_polygon3d(int points, int *pNx, int *pNy, int *pNz, int texture)
{
	// Plot a CONVEX polygon with 3 or more points/sides using one of two:
	// If texture >= 0, use 'texture'. Otherwise, use global variable 'color'
	// This is the same as fill_polygon() except depth coordinates and
	// texture number are passed as well
	// Note: This function uses the PolyBuffer
	// Note: Polygon is clipped against top, bottom, left and right edge.
	// WARNING: concave polygons WILL BE PLOTTED INCORRECTLY
	int index;
	int endpoint;
	int startpoint;
	int x1;
	int y1;
	int z1;
	int x2;
	int y2;
	int z2;
	int pNz_delta;

	char debug[256];

	int first_known_good_x;
	int first_known_good_y;
	int first_known_good = 0;

	int last_known_good_x;
	int last_known_good_y;
	int last_known_good = 0;

	reset_polybuffer();


	// Clip each leg against z=0
	for (index=0; index<points; index++)
	{
		startpoint = index;
		endpoint = index+1;
		if (endpoint >= points) { endpoint = endpoint - points; }

		if (pNz[startpoint] < 0 || pNz[endpoint] < 0) 
		{
			// This line must be clipped
			if (pNz[startpoint] >= 1)
			{
				// Start point is visible
				x1 = pNx[startpoint];
				y1 = pNy[startpoint];
				z1 = pNz[startpoint];

				// Interpolate second point
				pNz_delta = pNz[startpoint] - pNz[endpoint];
				if (pNz_delta == 0) { pNz_delta = 1; }
				x2 = pNx[endpoint] - ((pNz[endpoint] * (pNx[startpoint] - pNx[endpoint])) / pNz_delta);
				y2 = pNy[endpoint] - ((pNz[endpoint] * (pNy[startpoint] - pNy[endpoint])) / pNz_delta);
				z2 = 0;

				plot_leg(pNx[startpoint], pNy[startpoint], x2, y2);
				if (last_known_good == 1)
				{
					plot_leg(last_known_good_x, last_known_good_y, x2, y2);
				}
				last_known_good_x = x2;
				last_known_good_y = y2;
				last_known_good = 1;
				if (first_known_good == 0) 
				{
					first_known_good_x = x1;
					first_known_good_y = y1;
					first_known_good = 1;
				}
			}

			if (pNz[endpoint] >= 1) 
			{
				// End point is visible
				x2 = pNx[endpoint];
				y2 = pNy[endpoint];
				z2 = pNz[endpoint];

				// Interpolate first point
				pNz_delta = pNz[endpoint] - pNz[startpoint];
				if (pNz_delta == 0) { pNz_delta = 1; }
				x1 = pNx[startpoint] - ((pNz[startpoint] * (pNx[endpoint] - pNx[startpoint])) / pNz_delta);
				y1 = pNy[startpoint] - ((pNz[startpoint] * (pNy[endpoint] - pNy[startpoint])) / pNz_delta);
				z1 = 0;

				plot_leg(x1, y1, pNx[endpoint], pNy[endpoint]);
				if (last_known_good == 1)
				{
					plot_leg(x1, y1, last_known_good_x, last_known_good_y);
				}
				last_known_good_x = x2;
				last_known_good_y = y2;
				last_known_good = 1;
				if (first_known_good == 0) 
				{
					first_known_good_x = x1;
					first_known_good_y = y1;
					first_known_good = 1;
				}
			}
		}
		else
		{
			// Plot as-is
			plot_leg(pNx[startpoint], pNy[startpoint], pNx[endpoint], pNy[endpoint]);
			last_known_good_x = pNx[endpoint];
			last_known_good_y = pNy[endpoint];
			last_known_good = 1;
			if (first_known_good == 0) 
			{
				first_known_good_x = pNx[startpoint];
				first_known_good_y = pNy[startpoint];
				first_known_good = 1;
			}
		}
	}

	if (first_known_good == 1 && last_known_good == 1)
	{
		plot_leg(first_known_good_x, first_known_good_y, last_known_good_x, last_known_good_y);
	}

	if (texture == -1)
	{
		plot_polybuffer();
	}
	else
	{
		texture_polybuffer(texture);
	}

	/*
	sprintf(debug, "%i", pNz[0]+1);
	color = WHITE;
	plot_string(debug, pNx[0], pNy[0]);
	*/
}




int load_bitmap(char *bitmapname)
{
	// Load the specified bitmap into memory and return its index number
	// (If it has been loaded previously, simply return its index number)
	// If, for some reason, the bitmap can't be loaded, return -1

	// NOTE: This code doesn't actually load a bitmap yet, it
	// will just make a 64x64 pixel test pattern available
	FILE *filehandle;
	char filename[260];
//	char *path = "release\\resource\\bitmaps\\";
	char *path = "";
	char *filebuffer;
	unsigned int filesize;
	unsigned int bytesread;
	int bitmap = -1;
	int index;
	char loaded = 0;

	// For generating test pattern
	unsigned int *imagebuffer;
	unsigned int counter_x;
	unsigned int counter_y;
	int size;
	unsigned int rgb;


	// Start by looking through the bitmap list to see if this bitmap has been
	// loaded previously
	for (index=0; index<MAX_BITMAPS; index++)
	{
		if (strcmp(raster_bitmaps[index].filename, bitmapname) == 0)
		{
			return index;	// Function may exit here
		}
	}

	// The bitmap has not been loaded previously, so find an available slot
	for (index=0; index<MAX_BITMAPS; index++)
	{
		if (strcmp(raster_bitmaps[index].filename, "") == 0)
		{
			if (bitmap == -1) { bitmap = index; }
		}
	}

	// Did we find an available slot?
	if (bitmap >= 0)
	{
		// Yes. Now attempt to load the file into memory.
		filename[0] = 0;			// filename = ""
		strcat(filename, path);		// filename .= path
		strcat(filename, bitmapname);		// filename .= fontname
		filehandle = fopen(filename, "rb");
		if (filehandle != NULL)
		{
			// Determine the file size
			fseek(filehandle, 0, SEEK_END);		// Seek to the end...
			filesize = ftell(filehandle);		// ...to find out how big the file is
			rewind(filehandle);					// Rewind so we can read the file
			if (filesize > 0) 
			{
				// Allocate a buffer
				filebuffer = (unsigned char *) malloc(filesize);
				if (filebuffer != NULL) 
				{
					
						// GENERATE A TEST BITMAP -- REPLACE WITH ACTUAL IMAGE LOADING/DECODING CODE
						loaded = 1;
						strcpy(raster_bitmaps[bitmap].filename, filename);
						raster_bitmaps[bitmap].height = 64;
						raster_bitmaps[bitmap].width = 64;
						size = raster_bitmaps[bitmap].height * raster_bitmaps[bitmap].width * 4;
						imagebuffer = malloc(size);
						for (counter_y=0; counter_y<raster_bitmaps[bitmap].height; counter_y++)
						{
							for (counter_x=0; counter_x<raster_bitmaps[bitmap].width; counter_x++)
							{
								if (counter_x % 32 == counter_y % 32) { rgb = 0xFFFFFF; } else { rgb = 0x0000FF; }
								imagebuffer[(counter_y*raster_bitmaps[bitmap].width)+counter_x] = rgb;
							}						
							raster_bitmaps[bitmap].bitmap = imagebuffer;
						}						

				}
			}
			fclose(filehandle);
		}
	}

	if (loaded == 0) { bitmap = -1; }
	return bitmap;
}







// the basis function for a Bezier spline
float bezier_b(int i, float t) 
{
    switch (i) {
    case 0:
		return (1-t)*(1-t)*(1-t);
    case 1:
		return 3*t*(1-t)*(1-t);
    case 2:
		return 3*t*t*(1-t);
    case 3:
		return t*t*t;
    }
    return 0; //we only get here if an invalid i is specified
}
  
// the basis function for a B-spline
float bspline_b(int i, float t) 
{
    switch (i) {
    case 0:
		return ((-t + 2) * t - 1) * t / 2;
    case 1:
		return (((3 * t - 5) * t) * t + 2) / 2;
    case 2:
		return ((-3 * t + 4) * t + 1) * t / 2;
    case 3:
		return ((t - 1) * t * t) / 2;
    }
    return 0; //we only get here if an invalid i is specified
}
  
//evaluate a point on the bezier curve
vertex2d *bezier_p(vertex2d *ctrl_points, float t) 
{
	static vertex2d point;
	int i;
	point.x = 0;
	point.y = 0;
    for (i = 0; i<=3; i++){
		point.x += (int) (bezier_b(i, t) * ctrl_points[i].x);
		point.y += (int) (bezier_b(i, t) * ctrl_points[i].y);
    }
    return &point;
}


//evaluate a point on the B-spline curve
vertex2d *bspline_p(vertex2d *ctrl_points, float t) 
{
	static vertex2d point;
	int i;
	point.x = 0;
	point.y = 0;
    for (i = 0; i<=3; i++){
		point.x += (int) (bspline_b(i, t) * ctrl_points[i].x) + 0.5;
		point.y += (int) (bspline_b(i, t) * ctrl_points[i].y) + 0.5;
    }
    return &point;
}


void plot_curve(vertex2d *ctrl_points, int resolution)
{
	// Takes 4 control points 0-1-2-3, plots a B-spline between 1-2 
	vertex2d last;
	vertex2d *point;
	int i;
	for (i = 0; i <= resolution; i++) {
		point = bspline_p(ctrl_points, i/(float) resolution);
		if (i > 0) {
			plot_line(last.x, last.y, point->x, point->y);
		}
		last.x = point->x;
		last.y = point->y;
	}
}


void plot_spline(int points, int *pNx, int *pNy, int resolution) 
{
	// Takes a number of points and plots B-spline curves between them
	// forming an open continous curve passing through each point
	int index;
	vertex2d ctrl_points[4];

	for (index=1; index<points; index++)
	{
		if (index == 1) {					// Exception: first point
			ctrl_points[0].x = pNx[index-1]; ctrl_points[0].y = pNy[index-1]; 
			ctrl_points[1].x = pNx[index-1]; ctrl_points[1].y = pNy[index-1]; 
			ctrl_points[2].x = pNx[index+0]; ctrl_points[2].y = pNy[index+0]; 
			ctrl_points[3].x = pNx[index+1]; ctrl_points[3].y = pNy[index+1]; 
		} else if (index == points-1) {		// Exception: last point
			ctrl_points[0].x = pNx[index-2]; ctrl_points[0].y = pNy[index-2]; 
			ctrl_points[1].x = pNx[index-1]; ctrl_points[1].y = pNy[index-1]; 
			ctrl_points[2].x = pNx[index+0]; ctrl_points[2].y = pNy[index+0]; 
			ctrl_points[3].x = pNx[index+0]; ctrl_points[3].y = pNy[index+0]; 
		} else {							// Default
			ctrl_points[0].x = pNx[index-2]; ctrl_points[0].y = pNy[index-2]; 
			ctrl_points[1].x = pNx[index-1]; ctrl_points[1].y = pNy[index-1]; 
			ctrl_points[2].x = pNx[index+0]; ctrl_points[2].y = pNy[index+0]; 
			ctrl_points[3].x = pNx[index+1]; ctrl_points[3].y = pNy[index+1]; 
		}
		plot_curve(ctrl_points, resolution);
	}
}



vertex2d *get_curvepoint(vertex2d *ctrl_points, int resolution, int index)
{
	// Takes 4 control points 0-1-2-3, returns point no.'index' of 'resolution' 
	// between control points 1 and 2
	return bspline_p(ctrl_points, index/(float) resolution);
}


vertex2d *get_splinepoint(int points, int *pNx, int *pNy, int resolution, int index) 
{
	// Takes a number of points and calculates the position of point no.'index'
	// along that path
	vertex2d ctrl_points[4];
	int curve;

	curve = (index / resolution) + 1;
	index = index % resolution;

	if (curve == 1) {					// Exception: first point
		ctrl_points[0].x = pNx[curve-1]; ctrl_points[0].y = pNy[curve-1]; 
		ctrl_points[1].x = pNx[curve-1]; ctrl_points[1].y = pNy[curve-1]; 
		ctrl_points[2].x = pNx[curve+0]; ctrl_points[2].y = pNy[curve+0]; 
		ctrl_points[3].x = pNx[curve+1]; ctrl_points[3].y = pNy[curve+1]; 
	} else if (curve == points-1) {		// Exception: last point
		ctrl_points[0].x = pNx[curve-2]; ctrl_points[0].y = pNy[curve-2]; 
		ctrl_points[1].x = pNx[curve-1]; ctrl_points[1].y = pNy[curve-1]; 
		ctrl_points[2].x = pNx[curve+0]; ctrl_points[2].y = pNy[curve+0]; 
		ctrl_points[3].x = pNx[curve+0]; ctrl_points[3].y = pNy[curve+0]; 
	} else {							// Default
		ctrl_points[0].x = pNx[curve-2]; ctrl_points[0].y = pNy[curve-2]; 
		ctrl_points[1].x = pNx[curve-1]; ctrl_points[1].y = pNy[curve-1]; 
		ctrl_points[2].x = pNx[curve+0]; ctrl_points[2].y = pNy[curve+0]; 
		ctrl_points[3].x = pNx[curve+1]; ctrl_points[3].y = pNy[curve+1]; 
	}
	return get_curvepoint(ctrl_points, resolution, index);
}


// eoc.




/* EXAMPLES

		// Plot random triangles with white border
		for (index=0; index<200; index++)
        {
			color = fast_random(0xFFFFFF);
			p1x = fast_random(640);
			p1y = fast_random(480);
			p2x = fast_random(640);
			p2y = fast_random(480);
			p3x = fast_random(640);
			p3y = fast_random(480);
			fill_triangle(p1x, p1y, p2x, p2y, p3x, p3y);
			color = WHITE;
			plot_triangle(p1x, p1y, p2x, p2y, p3x, p3y);
		}


		// Plot random lines
		for (index=0; index<100; index++)
        {
			color = fast_random(0xFFFFFF);
			p1x = fast_random(640);
			p1y = fast_random(480);
			p2x = fast_random(640);
			p2y = fast_random(480);
			plot_line(p1x, p1y, p2x, p2y);
		}



		// Plot random rectangles with white border
		for (index=0; index<1; index++)
        {
			color = fast_random(0xFFFFFF);
			p1x = fast_random(640);
			p1y = fast_random(480);
			p2x = fast_random(640);
			p2y = fast_random(480);
			fill_rectangle(p1x, p1y, p2x, p2y);
			color = WHITE;
			plot_rectangle(p1x, p1y, p2x, p2y);
		}



		// Plot random circles with white border
		for (index=0; index<100; index++)
        {
			color = fast_random(0xFFFFFF);
			p1x = fast_random(620)+10;
			p1y = fast_random(460)+10;
			radius = fast_random(10);
			fill_circle(p1x, p1y, radius);
			color = WHITE;
			plot_circle(p1x, p1y, radius);
		}


	// You get the idea... :-)
*/



// eof.
